/*
 * JSVG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */


package com.jsevy.jsvg;

import com.jsevy.jxml.*;

import java.awt.Graphics2D;
import java.io.StringWriter;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

/**
 * Class representing a Scalable Vector Graphics document, which owns an svgGraphics on which drawing commands can be made. The document's
 * toSVGString method can then be called to generate the SVG text corresponding to the drawing done on the Graphics object.
 * 
 * The typical workflow is as follows:
 * 
 * <pre>
 * 
 * // Create an svg document and get its associated SVGGraphics instance
 * SVGDocument svgDocument = new SVGDocument(); 
 * SVGGraphics svgGraphics = svgDocument.getGraphics(); 
 * 
 * // Do drawing commands as on any other Graphics. If you have a paint(Graphics) method, 
 * // you can just use it with the SVGGraphics instance since it's a subclass of Graphics. 
 * graphics.setColor(Color.RED);  
 * graphics.setStroke(new BasicStroke(3));
 * graphics.drawLine(0, 0, 1000, 500); 
 * graphics.drawRect(1000, 500, 150, 150); 
 * graphics.drawRoundRect(20, 200, 130, 100, 20, 10); 
 * 
 * // Get the flat SVG output as a string - it's just XML text - and  save  in a file for use 
 * // with LibreOffice Draw or any other draw package that accepts flat SVG (.svg) files
 * String stringOutput = svgDocument.toSVGString(); 
 * String filePath = "path/to/file.svg"; 
 * FileWriter fileWriter = new FileWriter(filePath); 
 * fileWriter.write(svgText); 
 * fileWriter.flush(); 
 * fileWriter.close();
 * 
 * </pre>
 * 
 * @author jsevy
 *
 */
public class SVGDocument extends XMLDocument
{
    
    //private Vector<SVGEntity> entities;
    
    private SVGGraphics graphics;
    
    private SVGStyles styles;
    
    
    private double documentWidth = 10;
    private double documentHeight = 10;
    
    
    
    /**
     * Create a new SVGDocument
     */
    public SVGDocument()
    {
        super("svg");
        
        this.graphics = new SVGGraphics(this);
        
        //this.entities = new Vector<SVGEntity>();
        this.styles = new SVGStyles();
        
        populateXML();
    }
    
    
    
    
    /** Set the units for interpreting the dimension values in the SVG file.
     *
     *  @param unitsCode  SVG code indicating units:
     *      0 = Unitless
     *      1 = Inches
     *      2 = Feet
     *      3 = Miles
     *      4 = Millimeter
     *      5 = Centimeters
     *      6 = Meters
     *      7 = Kilometers
     *      8 = Microinches
     *      9 = Mils
     *      10 = Yards
     * 
     */
    public void setUnits(int unitsCode)
    {
        String[] unitsMatrix = 
            {
            /* 0 = */ "",
            /* 1 = */ "in",
            /* 2 = */ "ft",
            /* 3 = */ "mi",
            /* 4 = */ "mm",
            /* 5 = */ "cm",
            /* 6 = */ "m",
            /* 7 = */ "km",
            /* 8 = */ "uin",
            /* 9 = */ "mils",
            /* 10 = */ "yd",
        };
        
        SVGEntity.units = unitsMatrix[unitsCode];
    }
    
    
    
    /**
     * Set the document width and height, which equates to the overall viewBox size for the SVG drawing and which
     * will be added as an attribute to the top-level svg element. Note that the units used are those set with a call to setUnits.
     * 
     * @param documentWidth    Document/viewBox width, in units set by call to setUnits()
     * @param documentHeight   Document/viewBox height, in units set by call to setUnits()
     */
    public void setDocumentSize(double documentWidth, double documentHeight)
    {
        this.documentWidth = documentWidth;
        this.documentHeight = documentHeight;
    }
    
    
    
    /**
     * Set the scale factor for dimensions in the Java drawing - how many Java units/pixels
     * correspond to one unit in the SVG output. Note that the units can be set with a call to 
     * setUnits
     * 
     * @param javaUnitsPerUnit  How many Java units/pixels correspond to one SVG unit
     */
    public void setScaleFactor(double javaUnitsPerUnit)
    {
        SVGEntity.scaleFactor = javaUnitsPerUnit;
    }
    
    
    
    
    /**
     * Get the SVGGraphics associated with this document for use with standard Graphics drawing operations to generate
     * an svg text representation.
     * 
     * @return	The SVGGraphics associated with this document, on which standard Java Graphics drawing calls can be made
     */
    public SVGGraphics getGraphics()
    {
        return this.graphics;
    }
    
    
    /**
     * Get a graphics style corresponding to the current Graphics2D instance, and whether the shape
     * should be stroked and/or filled. Will return an existing corresponding style if there is one,
     * or create a new one and add it to the list of global styles if needed.
     * 
     * @param graphics  Java Graphics2D instance
     * @param stroked   True if shape is to be stroked with current color of supplied graphics instance
     * @param filled    True if shape is to be filled with current color of supplied graphics instance
     * @return          Corresponding SVGGraphicsStyle
     */
    protected SVGGraphicsStyle getGraphicsStyle(Graphics2D graphics, boolean stroked, boolean filled)
    {
        return styles.getGraphicsStyle(graphics, stroked, filled);
    }
    
    
    /**
     * Get a text style corresponding to the current Graphics2D instance. Will return an existing corresponding
     * style if there is one, or create a new one and add it to the list of global styles if needed.
     * 
     * @param graphics  Java Graphics2D instance
     * @return          Corresponding SVGTextStyle
     */
    protected SVGTextStyle getTextStyle(Graphics2D graphics)
    {
        return styles.getTextStyle(graphics);
    }
    
    
    private void populateXML()
    {
        //this.setRootElementName("svg");
        
        // add namespace attribute
        this.getDocumentElement().setAttribute("xmlns:svg", "http://www.w3.org/2000/svg");
    }
    
    
    
    /**
     * Return the SVG text associated with this SVG document. This includes the header and
     * entities populated with content generated by graphics calls on the associated SVGGraphics.
     * 
     * @return The SVG text associated with this document.
     * @throws TransformerException 
     */
    public String toSVGString()
    {
        // first, add viewbox to set user scale; wait until XML generation so the units and page size can be set if desired
        this.getDocumentElement().setAttribute("width", SVGObject.addUnits(documentWidth));
        this.getDocumentElement().setAttribute("height", SVGObject.addUnits(documentHeight));
        this.getDocumentElement().setAttribute("viewBox", "0 0 " + documentWidth + " " + documentHeight);
        
        DOMSource domSource = new DOMSource(this);
        StringWriter stringWriter = new StringWriter();
        
        try
        {
            TransformerFactory tf = TransformerFactory.newInstance();
            Transformer transformer = tf.newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "no");
            transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
            
            transformer.transform(domSource, new StreamResult(stringWriter));
        }
        catch (TransformerException e)
        {
            // really shouldn't ever happen...
            System.out.println("TransformerException while creating SVG XML: " + e.getMessage());
        }
        
        //xmlBuffer.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + "\n");
        //xmlBuffer.append(this.toXMLString(0, 2));
        
        StringBuffer xmlBuffer = stringWriter.getBuffer();
        return xmlBuffer.toString();
    }
    
    
    /**
     * Utility method used by the associated SVGGraphics object.
     * 
     * @param entity an svgEntity instance
     */
    public void addEntity(SVGEntity entity)
    {
        //entities.add(entity);
        this.getDocumentElement().appendChild(entity);
    }
    
}