
/*
 /*
 * JSVG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */


package com.jsevy.jxml;

import java.util.*;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.UserDataHandler;
import org.w3c.dom.Text;



/**
 * Class representing a general XML node. Can have children, attributes, and plain text associated with it.
 * 
 * @author jsevy
 *
 */
public abstract class XMLNode implements Node, Cloneable
{
	
	private String name;
	
	private StringBuffer valueBuffer;
	
	//protected String text;
	
	private Node parent;
	
	private String namespaceURI = null;
	
	private Document ownerDocument;
	
	private XMLNodeList children;
	
	private String prefix = null;
	
	
	
	


	
	
	/**
	 * Create a new empty anonymous (unnamed) XML node.
	 */
	public XMLNode()
	{
		this("");
	}
	
	
	/**
	 * Create an empty XML node with the specified name.
	 * @param name     Name for the node.
	 */
	public XMLNode(String name)
	{
		this(name, "");
	}
	
	
	/**
	 * Create an XML node with the specified name and containing the specified text.
	 * @param name     Name for the node.
	 * @param text     Text contained by the node.
	 */
	public XMLNode(String name, String text)
	{
		this.name = new String(name);
		this.valueBuffer = new StringBuffer(text);
		
		this.children = new XMLNodeList();
        
        this.parent = null;
        this.ownerDocument = null;
	}


    @Override
    public String getNodeName()
    {
        // name may be determined by subclass (e.g., comment, text, cdatasection nodes)
        if (this.prefix == null)
            return this.name;
        else
            return this.prefix + ":" + this.name;
    }
    
    
    /**
     * Note that this sets the non-prefixed part of the name - the local name
     * @param nodeName
     */
    public void setNodeName(String nodeName)
    {
        this.name = nodeName;
    }


    @Override
    public String getNodeValue() throws DOMException
    {
        // May well be null, depending on subclass
        if (this.valueBuffer != null)
            return this.valueBuffer.toString();
        else
            return null;
    }


    @Override
    public void setNodeValue(String nodeValue) throws DOMException
    {
        this.valueBuffer = new StringBuffer(nodeValue);
    }
    
    
    protected StringBuffer getValueBuffer()
    {
        return this.valueBuffer;
    }
    

    @Override
    public Node getParentNode()
    {
        return this.parent;
    }
    
    
    
    public void setParentNode(Node parent)
    {
        this.parent = parent;
    }


    @Override
    public NodeList getChildNodes()
    {
        return children;
    }
    
    
    public XMLNodeList getDescendants()
    {
        XMLNodeList nodeList = new XMLNodeList();
        
        // add this, then all children
        nodeList.add(this);
        
        for (int i = 0; i < this.getChildNodes().getLength(); i++)
        {
            nodeList.addAll(((XMLNode)this.getChildNodes().item(i)).getDescendants());
        }
        
        return nodeList;
    }


    @Override
    public Node getFirstChild()
    {
        if (children.getLength() > 0)
            return children.elementAt(0);
        else
            return null;
    }


    @Override
    public Node getLastChild()
    {
        int numChildren = children.getLength();
        
        if (numChildren > 0)
            return children.elementAt(numChildren - 1);
        else
            return null;
    }


    @Override
    public Node getPreviousSibling()
    {
        // use parent
        if (parent != null)
        {
            NodeList nodeList = parent.getChildNodes();
            
            for (int i = 1; i < nodeList.getLength(); i++)
            {
                Node node = nodeList.item(i);
                if (node == this)
                    return nodeList.item(i-1);
            }
            
            // get here only if we're the first child, or if in error we're not in the list
            return null;
            
        }
        else
        {
            return null;
        }
    }


    @Override
    public Node getNextSibling()
    {
     // use parent
        if (parent != null)
        {
            NodeList nodeList = parent.getChildNodes();
            
            for (int i = 0; i < nodeList.getLength() - 1; i++)
            {
                Node node = nodeList.item(i);
                if (node == this)
                    return nodeList.item(i+1);
            }
            
            // get here only if we're the last child (or if in error we're not in the list)
            return null;
            
        }
        else
        {
            return null;
        }
    }


    @Override
    public NamedNodeMap getAttributes()
    {
        // overridden by XMLElement, which is the only Node type with attributes
        return null;    
    }


    @Override
    public Document getOwnerDocument()
    {
        return this.ownerDocument;
    }
    
    
    public void setOwnerDocument(Document owner, boolean deep)
    {
        this.ownerDocument = owner;
        
        // do children recursively if deep and possible...
        if (deep)
        {
            for (int i = 0; i < children.getLength(); i++)
            {
                Node child = children.item(i);
                if (child instanceof XMLNode)
                    ((XMLNode)child).setOwnerDocument(owner, deep);
            }
        }
    }


    @Override
    public Node insertBefore(Node newChild, Node refChild) throws DOMException
    {
        int index = children.indexOf(refChild);
        if (index < 0)
            throw new DOMException(DOMException.NOT_FOUND_ERR, "No such child");
        
        // set parent, if possible... the XML interface spec doesn't have a way to do this
        if (newChild instanceof XMLNode)
        {
            // set this as parent
            ((XMLNode)newChild).setParentNode(this);
        }
        else
        {
            throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "Could not set parent when inserting child");
        }
        
        children.insertElementAt(newChild, index);
        
        return newChild;
    }
    
    

    public Node insertAfter(Node newChild, Node refChild) throws DOMException
    {
        int index = children.indexOf(refChild);
        if (index < 0)
            throw new DOMException(DOMException.NOT_FOUND_ERR, "No such child");
        
        // set parent, if possible... the XML interface spec doesn't have a way to do this
        if (newChild instanceof XMLNode)
        {
            // set this as parent
            ((XMLNode)newChild).setParentNode(this);
        }
        else
        {
            throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "Could not set parent when inserting child");
        }
        
        children.insertElementAt(newChild, index+1);
        
        return newChild;
    }


    @Override
    public Node replaceChild(Node newChild, Node oldChild) throws DOMException
    {
        int index = children.indexOf(oldChild);
        if (index < 0)
            throw new DOMException(DOMException.NOT_FOUND_ERR, "No such child");
        
        children.removeElementAt(index);
        
        // set parent, if possible... the XML interface spec doesn't have a way to do this
        if (newChild instanceof XMLNode)
        {
            // set this as parent
            ((XMLNode)newChild).setParentNode(this);
        }
        else
        {
            throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "Could not set parent when replacing child");
        }

        children.insertElementAt(newChild, index);
        
        return oldChild;
    }


    @Override
    public Node removeChild(Node oldChild) throws DOMException
    {
        if (children.remove(oldChild))
            return oldChild;
        else
            throw new DOMException(DOMException.NOT_FOUND_ERR, "No such child");
    }


    @Override
    public Node appendChild(Node newChild) throws DOMException
    {
        // add child to end of tree, removing it if it's already there before appending it
        children.remove(newChild);
        
        // set parent, if possible... the XML interface spec doesn't have a way to do this
        if (newChild instanceof XMLNode)
        {
            // set this as parent
            ((XMLNode)newChild).setParentNode(this);
        }
        else
        {
            throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "Could not set parent when appending child");
        }

        children.add(newChild);
        
        return newChild;
    }
    
    
    public Node prependChild(Node newChild) throws DOMException
    {
        // add child to head of child list, removing it if it's already there before prepending it
        children.remove(newChild);
        
        // set parent, if possible... the XML interface spec doesn't have a way to do this
        if (newChild instanceof XMLNode)
        {
            // set this as parent
            ((XMLNode)newChild).setParentNode(this);
        }
        else
        {
            throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "Could not set parent when prepending child");
        }

        // put it at head of child list
        children.insertElementAt(newChild, 0);
        
        return newChild;
    }


    @Override
    public boolean hasChildNodes()
    {
        return (children.size() > 0);
    }
    
    
    /**
     * Define this to catch pesky CloneNotSupported exception, which will never happen...
     */
    @Override
    public Object clone()
    {
        try
        {
            return super.clone();
        }
        catch (CloneNotSupportedException e)
        {
            // won't happen....
            e.printStackTrace();
            return null;
        }
    }


    
    @Override
    public Node cloneNode(boolean deep)
    {
        //this.getClass().getConstructor(String.class, String.class);
        //Node newElement = new XMLElement(this.name, this.value.toString());
        
        // clone method return the appropriate class for us...
        XMLNode newNode = (XMLNode)this.clone();
        
        // set parent to null; will be reset for child nodes when they're appended
        ((XMLNode)newNode).setParentNode(null);
        
        // create new empty kids list
        newNode.children = new XMLNodeList();
        
        if (deep)
        {
            // clone kids recursively
            for (int i = 0; i < children.getLength(); i++)
            {
                Node child = children.item(i);
                Node newChild = child.cloneNode(deep);
                newNode.appendChild(newChild);
            }
        }
        
        return newNode;
        
    }
    


    @Override
    public void normalize()
    {
        // just remove empty Text nodes and merge adjacent ones
        
        Node child = this.getFirstChild();
        while (child != null)
        {
            Node sibling = child.getNextSibling();
            
            if (child.getNodeType() == Node.TEXT_NODE)
            {
                // whack it if whitespace
                if (((Text)child).isElementContentWhitespace())
                {
                    this.removeChild(child);
                }
                else
                {
                    // merge into next sibling if it's also a text node
                    if ((sibling != null) && (sibling.getNodeType() == Node.TEXT_NODE))
                    {
                        // merge text
                        sibling.setNodeValue(child.getNodeValue() + sibling.getNodeValue());
                        
                        // remove this child
                        this.removeChild(child);
                    }
                }
            }
            else
            {
                // not a text node - normalize children recursively
                child.normalize();
            }
            
            // move to next sibling
            child = sibling;
        }
            
        
    }


    @Override
    public boolean isSupported(String feature, String version)
    {
        // the answer is "no"
        return false;
    }


    @Override
    public String getNamespaceURI()
    {
        return this.namespaceURI;
    }
    
    
    public void setNamespaceURI(String namespaceURI)
    {
        this.namespaceURI = namespaceURI;
    }


    @Override
    public String getPrefix()
    {
        // TODO Auto-generated method stub
        return this.prefix;
    }


    @Override
    public void setPrefix(String prefix) throws DOMException
    {
        // prefix only applicable to element and attribute nodes - null for all others
        if ((this.getNodeType() == Node.ELEMENT_NODE) || (this.getNodeType() == Node.ATTRIBUTE_NODE))
        {
            this.prefix = prefix;
        }
    }


    @Override
    public String getLocalName()
    {
        // TODO Auto-generated method stub
        return this.name;
    }


    @Override
    public boolean hasAttributes()
    {
        // overridden by XMLElement, which is only node type with attributes
        return false;
    }


    @Override
    public String getBaseURI()
    {
        // TODO Auto-generated method stub
        return null;
    }


    @Override
    public short compareDocumentPosition(Node other) throws DOMException
    {
        // same node, no bits set in mask
        if (this.isSameNode(other))
            return 0;
        
        // get parent lists - see if one is parent of the other
        XMLNodeList ancestors = new XMLNodeList();
        Node parent = this;
        while (parent != null)
        {
            ancestors.insertElementAt(parent, 0);
            parent = parent.getParentNode();
        }
        
        XMLNodeList otherAncestors = new XMLNodeList();
        parent = other;
        while (parent != null)
        {
            otherAncestors.insertElementAt(parent, 0);
            parent = parent.getParentNode();
        }
        
        
        // walk down the lists to find the last common ancestor
        int i = 0;
        int length = Math.min(ancestors.getLength(), otherAncestors.getLength());
        while ((i < length) && (ancestors.item(i).isSameNode(otherAncestors.item(i))))
        {
            i++;
        }
        i--;
        
        
        if (i < 0)
        {
            // no common ancestor; return disconnected plus implementation-specific, as per spec
            return (Node.DOCUMENT_POSITION_DISCONNECTED | Node.DOCUMENT_POSITION_IMPLEMENTATION_SPECIFIC);
        }
        else
        {
            // have last common ancestor; first see if it's one of our nodes
            Node lastCommonAncestor = ancestors.item(i);
            
            if (lastCommonAncestor.isSameNode(this))
            {
                // other is one of our kids - it is contained by us and follows us
                return (Node.DOCUMENT_POSITION_CONTAINED_BY | Node.DOCUMENT_POSITION_FOLLOWING);
            }
            
            if (lastCommonAncestor.isSameNode(other))
            {
                // we're one of his kids - it cotains us us and follows us
                return (Node.DOCUMENT_POSITION_CONTAINS | Node.DOCUMENT_POSITION_PRECEDING);
            }
            
            // neither is descendant of other; check order of kids of last common ancestor
            NodeList children = lastCommonAncestor.getChildNodes();
            for (int j = 0; j < children.getLength(); j++)
            {
                Node child = children.item(j);
                
                if (child.isSameNode(ancestors.item(i+1)))
                {
                    // our ancestor came first
                    return Node.DOCUMENT_POSITION_FOLLOWING;
                }
                
                if (child.isSameNode(otherAncestors.item(i+1)))
                {
                    // his ancestor came first
                    return Node.DOCUMENT_POSITION_PRECEDING;
                }
                
            }
            
            // really no way to get here... return disconnected to signal the issue...
            return (Node.DOCUMENT_POSITION_DISCONNECTED | Node.DOCUMENT_POSITION_IMPLEMENTATION_SPECIFIC);
            
        }
        
    }

    

    /*
    @Override
    public String getTextContent() throws DOMException
    {
        // return depends on subclass, so overridden
        return null;
    }
    */


    @Override
    public void setTextContent(String textContent) throws DOMException
    {
        // remove all children and replace with single XMLText node... hope this is what you wanted...
        this.children.clear();
        this.appendChild(new XMLText(textContent));
    }


    @Override
    public boolean isSameNode(Node other)
    {
        return (other == this);
    }


    @Override
    public String lookupPrefix(String namespaceURI)
    {
        // TODO Auto-generated method stub
        return null;
    }


    @Override
    public boolean isDefaultNamespace(String namespaceURI)
    {
        // TODO Auto-generated method stub
        return false;
    }


    @Override
    public String lookupNamespaceURI(String prefix)
    {
        // TODO Auto-generated method stub
        return null;
    }
    
    
    // need this so can do easy test on NamedNodeMap and NodeList
    public boolean equals(Object other)
    {
        if (other instanceof Node)
            return this.isEqualNode((Node)other);
        else
            return false;
    }


    /**
     * Two nodes are equal if and only if the following conditions are satisfied:
     *   - The two nodes are of the same type.
     *   - The following string attributes are equal: nodeName, localName, namespaceURI, prefix, nodeValue. 
     *     This is: they are both null, or they have the same length and are character for character identical.
     *   - The attributes NamedNodeMaps are equal. This is: they are both null, or they have the same length and 
     *     for each node that exists in one map there is a node that exists in the other map and is equal, although 
     *     not necessarily at the same index.
     *   - The childNodes NodeLists are equal. This is: they are both null, or they have the same length and contain
     *     equal nodes at the same index. Note that normalization can affect equality; to avoid this, nodes should
     *     be normalized before being compared.
     *   
     * For two DocumentType nodes to be equal, the following conditions must also be satisfied:
     *   - The following string attributes are equal: publicId, systemId, internalSubset.
     *   - The entities NamedNodeMaps are equal.
     *   - The notations NamedNodeMaps are equal.
     *
     * On the other hand, the following do not affect equality: the ownerDocument, baseURI, and parentNode attributes, 
     * the specified attribute for Attr nodes, the schemaTypeInfo attribute for Attr and Element nodes, the 
     * Text.isElementContentWhitespace attribute for Text nodes, as well as any user data or event listeners registered 
     * on the nodes. 
     */
    @Override
    public boolean isEqualNode(Node other)
    {
        // same node type
        if (this.getNodeType() != other.getNodeType())
            return false;
        
        // Objects.equals does the null test for us
        if (!(Objects.equals(this.getNodeName(), other.getNodeName())))
            return false;
        
        if (!(Objects.equals(this.getLocalName(), other.getLocalName())))
            return false;
        
        if (!(Objects.equals(this.getNamespaceURI(), other.getNamespaceURI())))
            return false;
        
        if (!(Objects.equals(this.getPrefix(), other.getPrefix())))
            return false;
        
        if (!(Objects.equals(this.getNodeValue(), other.getNodeValue())))
            return false;
        
        // now check attributes; really easy using Objects.equals method - will call the
        // equals() method for HashMap, which checks to see if the maps are equal by
        // calling equals() methods on everything involved.
        if (!(Objects.equals(this.getAttributes(), other.getAttributes())))
            return false;
        
        // now check children; again, really easy using Objects.equals method - will call the
        // equals() method for HashMap, which checks to see if the maps are equal by
        // calling equals() methods on everything involved.
        if (!(Objects.equals(this.getChildNodes(), other.getChildNodes())))
            return false;
        
        // for DocumentType nodes, a few more tests
        if (this.getNodeType() == Node.DOCUMENT_TYPE_NODE)
        {
            DocumentType thisDoc = (DocumentType)this;
            DocumentType otherDoc = (DocumentType)other;
            
            if (!(Objects.equals(thisDoc.getPublicId(), otherDoc.getPublicId())))
                return false;
            
            if (!(Objects.equals(thisDoc.getSystemId(), otherDoc.getSystemId())))
                return false;
            
            if (!(Objects.equals(thisDoc.getInternalSubset(), otherDoc.getInternalSubset())))
                return false;
            
            if (!(Objects.equals(thisDoc.getEntities(), otherDoc.getEntities())))
                return false;
            
            if (!(Objects.equals(thisDoc.getNotations(), otherDoc.getNotations())))
                return false;
            
        }
        
        // if we made it here, cleared all the specified hurdles - they're equal... whew...
        return true;
    }


    @Override
    public Object getFeature(String feature, String version)
    {
        // don't support features, whatever they are...
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "Features not supported.");
    }


    @Override
    public Object setUserData(String key, Object data, UserDataHandler handler)
    {
        // don't support user data at present
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "User data not supported.");
    }


    @Override
    public Object getUserData(String key)
    {
        // don't support user data at present
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "User data not supported.");
    }
    
    
    
    /*
      
    // Append the specified text to the text already contained in this node.
    // @param additionalText   Text to be added
    //
    public void appendText(String additionalText)
    {
        text += additionalText;
    }
    
    
    //
    // Set the text contained by this node to that specified.
    // @param newText   Replacement text
    //
    public void setText(String newText)
    {
        this.text = new String(newText);
    }
    
    
    //
    // Return the text contained in this node.
    // @return The node's text
    //
    public String getText()
    {
        return this.text.trim();
    }
    
    
    
    
    //
     * Find the first child with the specified name; return null if no such child exists. Note that
     * this considers only children, i.e., direct descendants; see findDescendant() for deeper search
     * in the tree of descendant nodes.
     * @param name     Name of the child desired
     * @return         Reference to the first child found with the specified name, or null if no such children.
    //
    public XMLNode findChild(String name)
    {
        for (int i = 0; i < children.size(); i++)
        {
            XMLNode child = (XMLNode)children.elementAt(i);
            
            if (child.name.equals(name))
            {
                return child;
            }
        }
        
        return null;
    }
    
    
    
    //
     * Find all descendants of this node which have the specified name. Does a complete search of
     * the tree of descendants, and returns a list with all corresponding descendants.
     * @param name     Name of descendants which are to be found
     * @return         Vector of descendant nodes which have the specified name; may be empty if no such nodes
    //
    public Vector<XMLNode> findDescendants(String name)
    {
        Vector<XMLNode> descendants = new Vector<XMLNode>();
        findDescendants(name, descendants);
        return descendants;
    }
    
    
    
    private void findDescendants(String name, Vector<XMLNode> descendants)
    {
        
        for (int i = 0; i < children.size(); i++)
        {
            XMLNode child = children.item(i);
            
            if (child.name.equals(name))
            {
                descendants.add(child);
            }
            
            child.findDescendants(name, descendants);
        }
    }
    
    
    //
     * Generate the XML text corresponding to this node and all of its descendants.
     * @return     The XML text for this node and its descendants
    //
    public String toXMLString()
    {
        StringBuilder result = new StringBuilder();
        
        result.append("<" + name);
        
        // add attributes
        for (int i = 0; i < attributes.size(); i++)
        {
            result.append(" " + (attributes.elementAt(i)).toXMLString());
        }
        
        result.append(">\n");
        
        if (!text.equals(""))
            result.append(text + "\n");
        
        for (int i = 0; i < children.size(); i++)
        {
            result.append((children.elementAt(i)).toXMLString());
        }
        
        result.append("</" + name + ">\n");
        
        return result.toString();
        
    }
    
    
    
    //
     * Generate XML text for this node and all of its descendants. Include extra spacing for 
     * hierarchical output formatting.
     * @param n        How many spaces to pre-pend for this node, when it is already in a hierarchical listing
     * @param deltaN   How many additional spaces to add for each level
     * @return         The hierarchically formatted XML text for this node and its descendants
    //
    public String toXMLString(int n, int deltaN)
    {
        StringBuilder result = new StringBuilder();
        
        String prefix = new String();
        for (int i = 0; i < n; i++)
            prefix += " ";
        
        String deltaPrefix = new String();
        for (int i = 0; i < deltaN; i++)
            deltaPrefix += " ";
        
        
        result.append(prefix + "<" + name);
        // add attributes
        for (int i = 0; i < attributes.size(); i++)
        {
            result.append(" " + (attributes.elementAt(i)).toXMLString());
        }
        result.append(">");
        
        result.append(prefix + deltaPrefix + text + "\n");
        
        for (int i = 0; i < children.size(); i++)
        {
            result.append((children.elementAt(i)).toXMLString(n + deltaN, deltaN));
        }
        
        result.append(prefix + "</" + name + ">\n");
        
        return result.toString();
        
    }
    
    
    public boolean equals(XMLNode otherNode)
    {
        // see if node type is same
        
        // check nodeName, localName, namespaceURI, prefix, nodeValue
    }
      
      
      
    */
	
	
}